package org.apache.xindice.server.components;

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * $Id: DBConduit.java,v 1.1.1.1 2001/12/06 19:33:56 bradford Exp $
 */

import org.apache.xindice.server.*;
import org.apache.xindice.util.*;
import java.io.*;
import java.sql.*;
import java.util.*;

/**
 * DBConduit provides a JDBC Connection pooling mechanism.
 * It also allows driver and connection information to be accessed via
 * centrally managed friendly alias names instead of hardcoded JDBC URLs.
 */

public final class DBConduit extends SimpleScriptComponent {
   private static final String ALIAS = "alias";
   private static final String NAME = "name";
   private static final String DRIVER = "driver";
   private static final String URL = "url";
   private static final String USERNAME = "username";
   private static final String PASSWORD = "password";
   private static final String DESCRIPTION = "description";
   private static final String CONNECTIONS = "connections";
   private static final String CREATE = "create";
   private static final String PING = "ping";

   private static final String[] EmptySet = new String[0];
   
   private Map aliases = new HashMap(); // ConnectionPool
   private Properties dbparams = new Properties();
   private String filename = null;
   private long pingCounter = 0;
   private ThreadGroup poolGroup = null;

   public void setConfig(Configuration config) throws XindiceException {
      super.setConfig(config);
      config.processChildren(ALIAS,
         new ConfigurationCallback() {
            public void process(Configuration cfg) {
               String name = cfg.getAttribute(NAME);
               dbparams.setProperty(name+".driver", cfg.getAttribute(DRIVER));
               dbparams.setProperty(name+".url", cfg.getAttribute(URL));
               dbparams.setProperty(name+".username", cfg.getAttribute(USERNAME));
               dbparams.setProperty(name+".password", cfg.getAttribute(PASSWORD));
               dbparams.setProperty(name+".description", cfg.getAttribute(DESCRIPTION));
               dbparams.setProperty(name+".connections", cfg.getAttribute(CONNECTIONS));
               dbparams.setProperty(name+".create", cfg.getAttribute(CREATE));
               dbparams.setProperty(name+".ping", cfg.getAttribute(PING));
            }
      });

      if ( poolGroup == null )
         poolGroup = new ThreadGroup(getName());
   }

   private synchronized ConnectionPool getConnectionPool(String alias) {
      ConnectionPool pool = (ConnectionPool)aliases.get(alias);
      if ( pool == null ) {
         String classname = (String)dbparams.get(alias+".driver");
         String url = (String)dbparams.get(alias+".url");
         String username = (String)dbparams.get(alias+".username");
         String password = (String)dbparams.get(alias+".password");
         String description = (String)dbparams.get(alias+".description");
         String ping = (String)dbparams.get(alias+".ping");
         int maxconns = StringUtilities.stringToInt((String)dbparams.get(alias+".connections"), 5);
         int create = StringUtilities.stringToInt((String)dbparams.get(alias+".create"), 0);
         if ( classname != null ) {
            pool = new ConnectionPool(alias, classname, url, username, password, description, ping, maxconns, create);
            aliases.put(alias, pool);
         }
      }
      return pool;
   }

   /**
    * listAliases retrieves a list of the aliases available to the
    * DBConduit
    *
    * @return A list of Aliases
    */
   public String[] listAliases() {
      String key;
      List l = new ArrayList();
      Iterator enum = dbparams.keySet().iterator();
      while ( enum.hasNext() ) {
         key = (String)enum.next();
         if ( key.endsWith(".driver") )
            l.add(key.substring(0, key.length()-7));
      }
      return (String[])l.toArray(EmptySet);
   }

   /**
    * getDescription returns a description for the specified alias.
    *
    * @param alias The alias name
    * @return The description
    */
   public String getDescription(String alias) {
      ConnectionPool pool = getConnectionPool(alias);
      return pool != null ? pool.getDescription()
                          : null;
   }

   /**
    * getDriver returns the driver class name for the specified
    * alias.
    *
    * @param alias The alias name
    * @return The driver class name
    */
   public String getDriver(String alias) {
      ConnectionPool pool = getConnectionPool(alias);
      return pool != null ? pool.getDriver()
                          : null;
   }

   /**
    * getPingStatement returns the SQL ping statement for the
    * specified alias.
    *
    * @param alias The alias name
    * @return The ping statement
    */
   public String getPingStatement(String alias) {
      ConnectionPool pool = getConnectionPool(alias);
      return pool != null ? pool.getPingStatement()
                          : null;
   }

   /**
    * getMaximumConnections returns the maximum allowable number of
    * open connections for the specified alias.
    *
    * @param alias The alias name
    * @return The maximum number of connections
    */
   public int getMaximumConnections(String alias) {
      ConnectionPool pool = getConnectionPool(alias);
      return pool != null ? pool.getMaximumConnections()
                          : -1;
   }

   /**
    * getCreatedConnections returns the number of open connections
    * for the specified alias.
    *
    * @param alias The alias name
    * @return The number of connections
    */
   public int getCreatedConnections(String alias) {
      ConnectionPool pool = getConnectionPool(alias);
      return pool != null ? pool.getCreatedConnections()
                          : -1;
   }

   /**
    * getTotalCreatedConnections returns the total number of
    * connections that have been opened during the lifetime of the
    * DBConduit.
    *
    * @param alias The alias name
    * @return The total number of connections
    */
   public int getTotalCreatedConnections(String alias) {
      ConnectionPool pool = getConnectionPool(alias);
      return pool != null ? pool.getTotalCreatedConnections()
                          : -1;
   }

   /**
    * getAvailableConnections returns the number of open
    * Connections available in the connection queue.
    *
    * @param alias The alias name
    * @return The number of available connections
    */
   public int getAvailableConnections(String alias) {
      ConnectionPool pool = getConnectionPool(alias);
      return pool != null ? pool.getAvailableConnections()
                          : -1;
   }

   /**
    * getConnection returns the next available Connection
    * from the connection queue or creates one if necessary.
    *
    * @param alias The alias name
    * @return A JDBC connection
    */
   public Connection getConnection(String alias) throws SQLException {
      ConnectionPool pool = getConnectionPool(alias);
      if ( pool != null )
         return pool.getConnection();
      else
         throw new SQLException("Alias "+alias+" not found in "+filename);
   }

   /**
    * putConnection returns a Connection to the connection
    * queue.
    *
    * @param alias The alias name
    * @param connection A JDBC connection
    */
   public void putConnection(String alias, Connection connection) {
      if ( connection == null )
         return;
      try {
         ConnectionPool pool = getConnectionPool(alias);
         if ( pool != null )
            pool.putConnection(connection);
      }
      catch ( Exception e ) {
      }
   }

   /**
    * removeConnection closes a connection and relieves the connection
    * queue of the task of maintaining it.
    *
    * @param alias The alias name
    * @param connection A JDBC connection
    */
   public void removeConnection(String alias, Connection connection) {
      if ( connection == null )
         return;
      try {
         ConnectionPool pool = getConnectionPool(alias);
         pool.removeConnection(connection);
      }
      catch ( Exception e ) {
      }
   }

   /**
    * executeQuery
    *
    * @param alias The alias name
    * @param sql The SQL to execute
    * @return The DBQueryResult object
    */
   public DBQueryResult executeQuery(String alias, String sql) throws SQLException {
      Connection connection = null;
      Statement statement = null;
      ResultSet resultset = null;
      DBQueryResult result = null;
      int rowsAffected = -1;
      try {
         connection = getConnection(alias);
         if ( connection != null ) {
            statement = connection.createStatement();
            if ( statement.execute(sql) )
               resultset = statement.getResultSet();
            else
               rowsAffected = statement.getUpdateCount();
            result = new DBQueryResult(this, alias, connection, statement, resultset, rowsAffected);
         }
      }
      catch ( SQLException e ) {
         if ( resultset != null )
            resultset.close();
         if ( statement != null )
            statement.close();
         if ( connection != null )
            removeConnection(alias, connection);
         throw e;
      }
      return result;
   }

   /**
    * ConnectionPool
    */

   private class ConnectionPool implements Runnable {
      private ObjectStack stack = new ObjectStack(); // Of Connection
      private ObjectQueue events = new ObjectQueue(); // of Event
      private String classname;
      private String name;
      private String url;
      private String username;
      private String password;
      private String description;
      private String ping;
      private int maxconns = 5;
      private int conns = 0;
      private int totalconns = 0;
      private Object mutex = new Object();
      private Thread manager;

      public ConnectionPool(String name, String classname, String url, String username, String password, String description, String ping, int maxconns, int create) {
         this.name = name;
         this.classname = classname;
         this.url = url;
         this.username = username;
         this.password = password != null ? password : "";
         this.description = description != null ? description : "";
         this.ping = ping;
         this.maxconns = maxconns;
         try {
            Driver driver = (Driver)Class.forName(classname).newInstance();
            DriverManager.registerDriver(driver);
         }
         catch ( Exception e ) {
            System.err.println("\u0007ERROR: JDBC driver "+classname+" not found for alias "+name);
         }
         if ( create > 0 ) {
            for ( int i = 0; i < create && i < maxconns; i++ )
               try {
                  stack.push(createConnection());
               }
               catch ( Exception e ) {
               }
         }
         manager = new Thread(poolGroup, this, "Connection Pool ("+name+")");
         manager.setDaemon(true);
         manager.start();
      }

      public Connection getConnection() throws SQLException {
         Connection connection;
         while ( true ) {
            connection = (Connection)stack.pop();
            if ( connection != null ) {
               if ( checkConnection(connection) )
                  return connection;
               else
                  conns--;
            }
            synchronized(mutex) {
               if ( conns < maxconns )
                  return createConnection();
               try {
                  mutex.wait();
               }
               catch ( Exception e ) {
               }
            }
         }
      }

      private Connection createConnection() throws SQLException {
         Connection connection;
         if ( username != null )
            connection = DriverManager.getConnection(url, username, password);
         else
            connection = DriverManager.getConnection(url);
         conns++;
         totalconns++;
         return connection;
      }

      private boolean checkConnection(Connection connection) {
         try {
            if ( connection.isClosed() )
               return false;
         }
         catch ( Exception e ) {
            return false;
         }
         if ( ping == null )
            return true;
         try {
            Statement statement = connection.createStatement();
            if ( ping.length() == 0 )
               statement.execute("SELECT "+(pingCounter++));
            else
               statement.execute(ping);
            statement.close();
            return true;
         }
         catch ( Exception e ) {
            return false;
         }
      }

      public void putConnection(Connection connection) {
         events.add(new Event(0, connection));
      }

      public void removeConnection(Connection connection) {
         events.add(new Event(1, connection));
      }

      public String getDescription() {
         return description;
      }

      public String getDriver() {
         return classname;
      }

      public String getPingStatement() {
         return ping;
      }

      public int getMaximumConnections() {
         return maxconns;
      }

      public int getCreatedConnections() {
         return conns;
      }

      public int getTotalCreatedConnections() {
         return totalconns;
      }

      public int getAvailableConnections() {
         return stack.size();
      }

      public void run() {
         Event event;
         boolean closed = true;
         while ( true ) {
            event = (Event)events.remove();
            if ( event != null ) {
               switch ( event.type ) {

                  case 0: // putConnection
                     closed = true;
                     try {
                        closed = event.connection.isClosed();
                     }
                     catch ( Exception e ) {
                     }
                     synchronized(mutex) {
                        if ( closed )
                           conns--;
                        else
                           stack.push(event.connection);
                        mutex.notify();
                     }
                     break;

                  case 1: // removeconnection
                     try {
                        event.connection.close();
                     }
                     catch ( Exception e ) {
                     }
                     synchronized(mutex) {
                        conns--;
                        mutex.notify();
                     }
                     break;

               }
               //Thread.yield();
            }
            else
               try {
                  Thread.sleep(100);
               }
               catch ( Exception e ) {
               }
         }
      }

      /**
       * Event
       */

      private class Event {
         public int type = 0;
         public Connection connection;

         public Event(int type, Connection connection) {
            this.type = type;
            this.connection = connection;
         }
      }
   }
}

