package org.apache.xindice.util;

/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xindice" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999-2001, The dbXML
 * Group, L.L.C., http://www.dbxmlgroup.com.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * $Id: LockManager.java,v 1.1.1.1 2001/12/06 19:33:58 bradford Exp $
 */

import java.util.*;

/**
 * LockManager manages resource locks.  A resource is any shared object that
 * can be represented as a long.  The LockManager should be constructed using
 * the maximum number of concurrent clients a specific resource might have.
 */

public final class LockManager {
   private int maxLocks = 0;
   private Map locks = new HashMap(); // Long to LockInfo

   public LockManager(int maxLocks) {
      this.maxLocks = maxLocks;
   }

   /**
    * acquireSharedLock acquires a shared lock on a resource.  Shared locks
    * are typically used for reads.
    */
   public void acquireSharedLock(long id) {
      acquireSharedLock(new Long(id));
   }

   /**
    * acquireSharedLock acquires a shared lock on a resource.  Shared locks
    * are typically used for reads.
    */
   public void acquireSharedLock(Object key) {
      LockInfo info;
      synchronized(this) {
         info = (LockInfo)locks.get(key);
         if ( info == null ) {
            info = new LockInfo();
            locks.put(key, info);
         }
      }
      synchronized(info) {
         while ( info.count >= maxLocks ) {
            try {
               info.wait();
            }
            catch ( Exception e ) {
            }
         }
         info.count++;
      }
   }

   /**
    * releaseSharedLock releases a shared lock on a resource.  Shared locks
    * are typically used for reads.
    */
   public void releaseSharedLock(long id) {
      releaseSharedLock(new Long(id));
   }

   /**
    * releaseSharedLock releases a shared lock on a resource.  Shared locks
    * are typically used for reads.
    */
   public void releaseSharedLock(Object key) {
      LockInfo info;
      synchronized(this) {
         info = (LockInfo)locks.get(key);
      }
      if ( info == null )
         return;
      synchronized(info) {
         info.count--;
         info.notify();
         if ( info.count == 0 ) {
            synchronized(this) {
               locks.remove(key);
            }
         }
      }
   }

   /**
    * acquireExclusiveLock acquires an exclusive lock on a resource.  Exclusive
    * locks are typically used for writes.
    */
   public void acquireExclusiveLock(long id) {
      acquireExclusiveLock(new Long(id));
   }

   /**
    * acquireExclusiveLock acquires an exclusive lock on a resource.  Exclusive
    * locks are typically used for writes.
    */
   public void acquireExclusiveLock(Object key) {
      LockInfo info;
      synchronized(this) {
         info = (LockInfo)locks.get(key);
         if ( info == null ) {
            info = new LockInfo();
            locks.put(key, info);
         }
      }
      synchronized(info) {
         int total = 0;
         int stolen = 0;
         while ( total < maxLocks && info.count >= maxLocks ) {
            try {
               info.wait();
            }
            catch ( Exception e ) {
            }
            stolen = maxLocks-info.count;
            total += stolen;
            info.count += stolen;
         }
      }
   }

   /**
    * releaseExclusiveLock releases an exclusive lock on a resource.  Exclusive
    * locks are typically used for writes.
    */
   public void releaseExclusiveLock(long id) {
      releaseExclusiveLock(new Long(id));
   }

   /**
    * releaseExclusiveLock releases an exclusive lock on a resource.  Exclusive
    * locks are typically used for writes.
    */
   public void releaseExclusiveLock(Object key) {
      LockInfo info;
      synchronized(this) {
         info = (LockInfo)locks.get(key);
      }
      if ( info == null )
         return;
      synchronized(info) {
         info.count = 0;
         info.notify();
         if ( info.count == 0 ) {
            synchronized(this) {
               locks.remove(key);
            }
         }
      }
   }


   /**
    * LockInfo is just an int wrapper.
    */
   private class LockInfo {
      public int count = 0;
   }
}
